from .graphi_types import Types
import graphene
from django.db.utils import IntegrityError
from graphene_file_upload.scalars import Upload
from django.db import models


class MutationsMixin(Types, graphene.ObjectType):
    @classmethod
    def map_field_to_graphene_type(cls, field_name: str, model):
        if field_name.endswith("_pk"):
            return graphene.String()

        if field_name.endswith("_pks"):
            return graphene.List(graphene.String)

        field = model._meta.get_field(field_name)
        field_type = field.get_internal_type()
        type_mapping = {
            'CharField': graphene.String(),
            'IntegerField': graphene.Int(),
            'BooleanField': graphene.Boolean(),
            'DateField': graphene.String(),
            'DateTimeField': graphene.String(),
            'DecimalField': graphene.Float(),
            'EmailField': graphene.String(),
            'FileField': Upload(),
            'FloatField': graphene.Float(),
            'ImageField': Upload(),
            'PositiveIntegerField': graphene.Int(),
            'PositiveSmallIntegerField': graphene.Int(),
            'SlugField': graphene.String(),
            'SmallIntegerField': graphene.Int(),
            'TextField': graphene.String(),
            'TimeField': graphene.String(),
            'UUIDField': graphene.String(),
            'BigIntegerField': graphene.Int(),
            'BinaryField': graphene.String(),
            'PositiveBigIntegerField': graphene.Int(),
            'PositiveDecimalField': graphene.Float(),
            'URLField': graphene.String(),
            'AutoField': graphene.Int(),
            'BigAutoField': graphene.Int(),
            'ForeignKey': graphene.String(),
        }
        return type_mapping.get(field_type)

    @classmethod
    def get_mutation_fields(cls, model):
        new_fields = []
        excluded_fields = ['created_at', 'updated_at', 'created_by', 'updated_by']

        for field_name in cls.get_fields(model):
            field = model._meta.get_field(field_name)

            if field_name in excluded_fields:
                continue

            if hasattr(field, 'primary_key'):
                if field.primary_key:
                    continue
            
            if not isinstance(field, models.Field):
                continue

            if field.get_internal_type() in ['ForeignKey', 'OneToOneField']:
                new_fields.append(f"{field_name}_pk")
            elif field.get_internal_type() == 'ManyToManyField':
                new_fields.append(f"{field_name}_pks")
            else:
                new_fields.append(field_name)

        return new_fields

    @classmethod
    def generate_input_type(cls, model):
        fields = cls.get_mutation_fields(model)
        return type(
            f"{model.__name__}InputType",
            (graphene.InputObjectType, ),
            {
                field: cls.map_field_to_graphene_type(field, model) for field in fields
            }
        )

    @classmethod
    def resolve_related_objects_from_input(cls, model, _input: dict):
        related_objects = {}
        for input_field in _input:
            if input_field.endswith('_pk'):
                field_name = input_field.replace('_pk', '')
                related_model = model._meta.get_field(field_name).related_model
                try:
                    related_objects[field_name] = related_model.objects.get(pk=_input[input_field])
                except related_model.DoesNotExist:
                    raise Exception(f'{related_model.__name__} with pk {_input[input_field]} does not exist.')
            elif input_field.endswith('_pks'):
                field_name = input_field.replace('_pks', '')
                related_model = model._meta.get_field(field_name).related_model
                related_objects[field_name] = related_model.objects.filter(pk__in=_input[input_field])
            else:
                related_objects[input_field] = _input[input_field]
        return related_objects

    @classmethod
    def resolve_many_to_many(cls, model, _input: dict):
        related_objects = {}
        _input_copy = _input.copy()
        for input_field in _input:
            if input_field.endswith('_pks'):
                field_name = input_field.replace('_pks', '')
                related_model = model._meta.get_field(field_name).related_model
                related_objects[field_name] = related_model.objects.filter(pk__in=_input[input_field])
                del _input_copy[input_field]

        return related_objects, _input_copy

    @classmethod
    def check_permission(cls, user, model):
        if not hasattr(model, 'graphql_permissions'):
            return True

        if not model.graphql_permissions:
            return True

        if not user.is_authenticated:
            raise Exception('Permission Denied')

        if not user.has_perms(model.graphql_permissions):
            raise Exception('Permission Denied')
        
        return True





from graphene_django import DjangoObjectType
from django.apps import apps

class Types:
    django_object_types = dict()

    objects_generated = False

    @classmethod
    def get_or_generate_django_object_type(cls, model):
        if model not in cls.django_object_types:
            cls.django_object_types[model] = cls.generate_model_type(model)
        return cls.django_object_types[model]

    @classmethod
    def get_fields(cls, model):
        excluded_fields = []
        if hasattr(model, "graphql_fields") and hasattr(
            model, "graphql_exclude_fields"
        ):
            raise Exception(
                f"please remove graphql_fields or graphql_exclude_fields from {model}"
            )

        if hasattr(model, "graphql_fields"):
            if not isinstance(model.graphql_fields, (list, tuple)):
                raise Exception(f"graphql_fields must be iterable on model {model}")
            return model.graphql_fields

        if hasattr(model, "graphql_exclude_fields"):
            if not isinstance(model.graphql_exclude_fields, (list, tuple)):
                raise Exception(
                    f"graphql_exclude_fields must be iterable on model {model}"
                )
            excluded_fields = model.graphql_exclude_fields
        fields = []
        for field in model._meta.get_fields():
            if field.name not in excluded_fields and hasattr(model, field.name):
                try:
                    if getattr(model, field.name).reverse:
                        continue
                except AttributeError:
                    pass
                fields.append(field.name)
        return fields

    @classmethod
    def generate_model_type(cls, model):
        if model.__name__ == "Employee":
            print('-'*10)
            print("model:", model)
            print(cls.get_fields(model))
        model_type = type(
            f"{model.__name__.lower()}",
            (DjangoObjectType,),
            {
                "Meta": type(
                    "Meta",
                    (),
                    {"model": model, "fields": cls.get_fields(model)},
                ),
            },
        )
        return model_type
    
    @classmethod
    def find_all_models(cls, _apps: list):
        all_apps = [apps.get_app_config(app) for app in _apps]
        for app in all_apps:
            models = app.get_models()
            for model in models:
                if hasattr(model, "graphql_exclude"):
                    if model.graphql_exclude:
                        continue
                yield model
